/***********************************************
* Hakan Ceylan
* PhD Student / Research Assistant
* Dept. of Computer Science
* University of North Texas
* Denton, TX, 76203, USA
*
* Email: hakan@unt.edu
*
************************************************/

#include "bplustree.h"

/******************************************************
************* ABSTRACT BASE NODE **********************
******************************************************/
int Node :: binarySearch( const Key& key, int start, int end ) const
{
	int middle = start + ( (end - start) / 2 );

	if( start == middle )
	{
		//if( this -> isLeafNode() && key == keys[middle] )
		//		return end;

		if( key < keys[middle] )
			return middle;
		else
			return end;
	}

	else if( key < keys[middle] )
		return binarySearch( key, start, middle );
	else
		return binarySearch( key, middle, end );
}

int Node :: findPosition( const Key& key ) const
{
	//may throw an exception if keys==NULL

	if( currentSize == 0 )
		return 0;
	return binarySearch( key, 0, currentSize );

	//Linear Search - Old Implementation
	/*
	int i;
	for( i = 0; i < currentSize; i++ )
	{
		if( key < keys[i] )
			break;
	}

	return i;
	*/
}

void Node :: shiftRight( int position )
{
	//may throw an exception if the extra allocated space is also full (currentSize==order*2+1)
	for( int i = currentSize-1; i >= position; i-- )
	{
		keys[i+1] = keys[i];
		if( locations != NULL )
			locations[i+2] = locations[i+1];
		if( pointers != NULL )
			pointers[i+2] = pointers[i+1];
	}
}

void Node :: addKey( const Key& key, Node* node )
{
	int position = findPosition( key );
	shiftRight( position );
	keys[position] = key;
	pointers[position+1] = node;
	currentSize++;
}

void Node :: addKey( const Key& key, DiskLocation location )
{
	int position = findPosition( key );
	shiftRight( position );
	keys[position] = key;
	locations[position+1] = location;
	currentSize++;
}

Node* Node :: split( const Key& key, Node* pointer, Key& bumpKey )
{
	addKey( key, pointer );

	Node* newNode = new Node( order, false );
	bumpKey = keys[currentSize/2];
	//Do not copy the key to be bumped
	newNode -> pointers[0] = pointers[currentSize/2+1];
	for( int i = currentSize/2+1; i < currentSize; i++ )
		newNode -> addKey( keys[i], pointers[i+1] );

	currentSize /= 2;

	return newNode;
}

Node* Node :: split( const Key& key, DiskLocation location, Key& bumpKey )
{
	addKey( key, location );

	Node* newNode = new Node( order, true );
	bumpKey = keys[currentSize/2];
	//Do not copy the key to be bumped
	newNode -> locations[0] = locations[currentSize/2+1];
	for( int i = currentSize/2+1; i < currentSize; i++ )
		newNode -> addKey( keys[i], locations[i+1] );

	currentSize /= 2;

	return newNode;
}

void Node :: print( ostream& os ) const
{
	os << "Order: " << order << endl;
	os << "Current Size: " << currentSize << endl;
	os << "Keys:" << endl;
	for( int i = 0; i < currentSize; i++ )
		os << keys[i] << endl;
}


/******************************************************
************* B+ Tree *********************************
******************************************************/
void BPlusTree :: eraseAll( Node* node )
{
	if( node == NULL )
		return;

	if( node -> isLeafNode() )
	{
		delete node;
		return;
	}

	for( int i = 0; i <= node -> currentSize; i++ )
		eraseAll( node -> pointers[i] );
	delete node;
}

void BPlusTree :: print( ostream& os, const Node* node, int level ) const
{
	os << "Level: " << level << endl;
	os << *node << endl;

	if( !(node -> isLeafNode()) )
		for( int i = 0; i <= node -> currentSize; i++ )
			print( os, node -> pointers[i], level + 1 );
}

void BPlusTree :: findLeafNode( const Key& key, Node* node, vector<Node*>& path ) const
{
	if( node -> isLeafNode() )
	{
		path.push_back( node );
		return;
	}

	for( int i = 0; i < node -> currentSize; i++ )
	{
		if( key < node -> keys[i] )
		{
			path.push_back( node );
			findLeafNode( key, node -> pointers[i], path );
			return;
		}
	}
	path.push_back( node );
	findLeafNode( key, node -> pointers[node->currentSize], path );
}

void BPlusTree :: bumpUp( const Key& bumpKey, Node* node, vector<Node*>& path )
{
	if( path.size() == 0 )
	{
		//bump up came from root, increasing depth
		Node *newRoot = new Node( order, false );
		newRoot -> pointers[0] = root;
		newRoot -> addKey( bumpKey, node );
		depth++;
		root = newRoot;
		return;
	}

	Node *currentNode = path.back();
	path.pop_back();
	if( currentNode -> isFull() )
	{
		Key newBumpKey;
		Node *newInternalNode = currentNode -> split( bumpKey, node, newBumpKey );
		bumpUp( newBumpKey, newInternalNode, path );
	}
	else
	{
		currentNode -> addKey( bumpKey, node );
	}
}

void BPlusTree :: insert( const Key& key, DiskLocation location )
{
	if( root == NULL )
	{
		root = new Node( order, true );
		//root -> addKey( key, location );
		root -> locations[0] = location;
		return;
	}

	vector<Node*> path( 0 );
	findLeafNode( key, root, path );
	Node *leafNode = path.back();
	path.pop_back();
	if( leafNode -> isFull() )
	{
		Key bumpKey;
		Node *newLeafNode = leafNode -> split( key, location, bumpKey );
		bumpUp( bumpKey, newLeafNode, path );
	}
	else
	{
		leafNode -> addKey( key, location );
	}
}

void BPlusTree :: build( const string& file )
{
	if( root != NULL )
	{
		eraseAll( root );
		root = NULL;
	}

	ifstream fo( file.c_str() );

	int count = 0;
	while( !fo.eof() )
	{
		Key key;
		DiskLocation location;
		try
		{
			if( ( count % bucketFactor ) == 0 )
			{
				location = readRecord( fo, key );
				insert( key, location );
			}
			else
				fo >> sampleRecord;

		}
		catch( const char* p )
		{
			if( fo.fail() )
				break;
			throw( p );
		}

		if( fo.fail() )
			break;
		count++;
	}
	fo.close();
	cout << "\t\tNumber of records inserted: " << count << endl;
}

DiskLocation BPlusTree :: writeHelper( ostream& os, Node* node ) const
{
	if( node -> isLeafNode() )
	{
		//write leaf node
		//return the pointer
		DiskLocation location = os.tellp();
		os << node -> currentSize << " L" << endl;
		for( int i = 0; i < node -> currentSize; i++ )
			os << node -> keys[i] << ' ' << node -> locations[i] << endl;
		os << node -> locations[node -> currentSize] << endl;
		return location;
	}

	node -> locations = new DiskLocation[node -> currentSize+1];
	for( int i = 0; i <= node -> currentSize; i++ )
		node -> locations[i] = writeHelper( os, node -> pointers[i] );

	//write internal node
	//return the pointer
	DiskLocation location = os.tellp();
	os << node -> currentSize << " I" << endl;
	for( int i = 0; i < node -> currentSize; i++ )
		os << node -> keys[i] << ' ' << node -> locations[i] << endl;
	os << node -> locations[node -> currentSize] << endl;

	return location;
}

DiskLocation BPlusTree :: write( const string& file ) const
{
	if( root == NULL )
		throw( "Nothing to write!" );

	ofstream fo( file.c_str() );
	DiskLocation location = writeHelper( fo, root );
	fo.close();

	return location;
}

Node* BPlusTree :: readNode( istream& is ) const
{
	int numOfKeys;
	char typeOfNode;
	is >> numOfKeys >> typeOfNode;

	Node* node;
	if( typeOfNode == 'L' )
	{
		node = new Node( order, true );
	}
	else if( typeOfNode == 'I' )
	{
		node = new Node( order, false );
		node -> locations = new DiskLocation[numOfKeys+1];
	}
	else
	{
		throw( "Type of node is not understood!" );
	}

	node -> currentSize = numOfKeys;
	for( int i = 0; i < numOfKeys; i++ )
	{
		Key key( sampleRecord.numOfWords );
		is >> key;

		//DiskLocation loc;
		streamoff loc;
		is >> loc;

		node -> keys[i] = key;
		node -> locations[i] = loc;
	}

	//DiskLocation loc;
	streamoff loc;
	is >> loc;
	node -> locations[numOfKeys] = loc;

	return node;
}

Node* BPlusTree :: readLevels( istream& is, DiskLocation location, int levelsToRead )
{
	if( levelsToRead == 0 )
		return NULL;

	is.seekg( location );

	Node* node = readNode( is );
	if( !(node -> isLeafNode()) )
	{
		for( int i = 0; i <= node -> currentSize; i++ )
		{
			node -> pointers[i] = readLevels( is, node -> locations[i], levelsToRead-1 );
		}
	}

	return node;
}

void BPlusTree :: read( const string& nodesFile, const string& recFile, DiskLocation rootLocation, int levelsToRead )
{
	if( root != NULL )
		eraseAll( root );

	indexFile = nodesFile;
	recordFile = recFile;

	ifstream fo( indexFile.c_str() );
	root = readLevels( fo, rootLocation, levelsToRead );
	depth = levelsToRead - 1;
}

DiskLocation BPlusTree :: readRecord( istream& is, Key &key )
{
	DiskLocation position = is.tellg();
	try
	{
		is >> sampleRecord;
		keyMap -> getKey( sampleRecord, indexPosition, key );
	}
	catch( const char * p )
	{
		throw( p );
	}

	return position;
}

void BPlusTree :: readRecord( DiskLocation location, Record& record ) const
{
	ifstream fo( recordFile.c_str() );
	fo.seekg( location );

	record = sampleRecord;
	fo >> record;
	fo.close();
}

DiskLocation BPlusTree :: findBucket( const Key& key, const Node* node ) const
{
	int pos = node -> findPosition( key );

	if( node -> isLeafNode() )
	{
		return node -> locations[pos];
	}
	else
	{
		if( node -> pointers[pos] )
			return findBucket( key, node -> pointers[pos] );
		else
		{
			ifstream fo( indexFile.c_str() );
			fo.seekg( node -> locations[pos] );
			Node* n = readNode( fo );
			if( !( n -> isLeafNode() ) )
			{
				for( int i = 0; i <= n -> currentSize; i++ )
					n -> pointers[i] = NULL;
			}
			DiskLocation location = findBucket( key, n );
			delete n;
			return location;
		}
	}
}

DiskLocation BPlusTree :: query( const Key& key ) const
{
	if( root == NULL )
		throw( "The tree is empty!" );

	return findBucket( key, root );
}
