/***********************************************
* Hakan Ceylan
* PhD Student / Research Assistant
* Dept. of Computer Science
* University of North Texas
* Denton, TX, 76203, USA
*
* Email: hakan@unt.edu
*
************************************************/

#ifndef _BPLUSTREE_H_
#define _BPLUSTREE_H_

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>

#include "key.h"
#include "record.h"
#include "keyMap.h"

using namespace std;

#define DEFAULT_ORDER	100
#define DEFAULT_BUCKET_FACTOR	100

typedef streampos DiskLocation;


/******************************************************
************* ABSTRACT BASE NODE **********************
******************************************************/
class Node
{
	private:
		int order;
		const int maxSize;

		int binarySearch( const Key&, int, int ) const;

	protected:
		void print( ostream& ) const;

	public:
		int currentSize;
		Key* keys;
		DiskLocation* locations;
		Node** pointers;

		Node() : currentSize( 0 ), order( DEFAULT_ORDER ), maxSize( order * 2 ), locations( NULL )
		{
			keys = new Key[maxSize+1];
			pointers = new Node*[maxSize+2];
		}
		Node( int o, bool isLeaf ) : currentSize( 0 ), order( o ), maxSize( order * 2 ), locations( NULL ), pointers( NULL )
		{
			keys = new Key[maxSize+1];
			if( isLeaf )
				locations = new DiskLocation[maxSize+2];
			else
				pointers = new Node*[maxSize+2];
		}
		~Node()
		{
			if( keys )
				delete[] keys;
			if( locations )
				delete[] locations;
			if( pointers )
				delete[] pointers;
		}

		bool isLeafNode() const
		{
			return ( pointers == NULL );
		}

		bool isFull() const
		{
			return (currentSize == maxSize);
		}

		int findPosition( const Key& ) const;
		void shiftRight( int );
		void addKey( const Key&, Node* );
		void addKey( const Key&, DiskLocation );
		Node* split( const Key&, Node*, Key& );
		Node* split( const Key&, DiskLocation, Key& );

		friend ostream& operator<<( ostream& os, const Node& node )
		{
			node.print( os );
			return os;
		}
};
	
/******************************************************
************* B+ Tree *********************************
******************************************************/
class BPlusTree
{
	private:
		Node* root;
		int depth;
		int order;
		int bucketFactor;
		int indexPosition;
		Record sampleRecord;
		const KeyMap* keyMap;

		string recordFile;
		string indexFile;

		void eraseAll( Node* );
		void print( ostream& os, const Node* node, int level ) const;
		void findLeafNode( const Key&, Node*, vector<Node*>& ) const;
		void bumpUp( const Key&, Node*, vector<Node*>& );
		void insert( const Key&, DiskLocation );
		DiskLocation writeHelper( ostream&, Node* ) const;

		Node* readNode( istream& ) const;
		Node* readLevels( istream&, DiskLocation, int );
		DiskLocation readRecord( istream&, Key& );
		void readRecord( DiskLocation, Record& ) const;
		DiskLocation findBucket( const Key&, const Node* ) const;

	public:
		BPlusTree() : root( NULL ), depth( 0 ), order( DEFAULT_ORDER ), bucketFactor( DEFAULT_BUCKET_FACTOR ), keyMap( NULL ), recordFile( "" ), indexFile( "" )
		{
		}
		BPlusTree( const Record& rec, int ip, const KeyMap* km, int bkfr = DEFAULT_BUCKET_FACTOR, int o = DEFAULT_ORDER ) : root( NULL ), depth( 0 ), sampleRecord( rec ), indexPosition( ip ), keyMap( km ), bucketFactor( bkfr ), order( o ), recordFile( "" ), indexFile( "" )
		{
		}
		~BPlusTree()
		{
			eraseAll( root );
		}

		int getDepth() const
		{
			return depth;
		}

		void build( const string& );
		DiskLocation write( const string& ) const;

		void read( const string&, const string&, DiskLocation, int );
		DiskLocation query( const Key& ) const;

		friend ostream& operator<<( ostream& os, const BPlusTree& tree )
		{
			tree.print( os, tree.root, 0 );
			return os;
		}
};

#endif
