/***********************************************
* Driver for building indices
*
* Hakan Ceylan
* PhD Student / Research Assistant
* Dept. of Computer Science
* University of North Texas
* Denton, TX, 76203, USA
*
* Email: hakan@unt.edu
*
************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

#include <ctime>
#include <cstdlib>

#include "key.h"
#include "record.h"
#include "keyMap.h"
#include "bplustree.h"

using namespace std;

void index( const int ngramSize, const int indexPosition, const int bucketFactor, const int order, const KeyMap* keyMap, const string& recordDir, const string& indexDir )
{
	const string rootIndexFileName = "roots";
	const string fileExtension = ".index";

	const Record sampleRecord( ngramSize );

	ostringstream oss;
	oss << ngramSize;

	const string idxFilePath = recordDir + oss.str() + "gm.idx";
	ifstream idxfo( idxFilePath.c_str() );

	ofstream fo( (indexDir + rootIndexFileName + fileExtension).c_str() );
	while( !idxfo.eof() )
	{
		string fileName, word;
		vector<string> firstNgram;
		idxfo >> fileName;
		for( int i = 0; i < ngramSize; i++ )
		{
			idxfo >> word;
			firstNgram.push_back( word );
		}

		if( idxfo.fail() )
			break;

		fileName.resize( fileName.length() - 3 );

		Key key( ngramSize );
		try
		{
			keyMap -> getKey( firstNgram, indexPosition, key );
		}
		catch( const char* p )
		{
			cout << "The first record has word \"" << p << "\" which is not found in the vocabulary!" << endl;
		}

		fo << recordDir + fileName << ' ' << indexDir + fileName + fileExtension << ' ' << indexPosition << ' ' << bucketFactor << ' ' << order << ' ' << key << ' ';

		cout << endl << "Indexing File \"" << recordDir + fileName << "\"..." << endl;
		time_t start = time( NULL );

		BPlusTree tree( sampleRecord, indexPosition, keyMap, bucketFactor, order );

		try
		{
			cout << "\tReading records and building B+ Tree..." << endl;
			tree.build( recordDir + fileName );
			cout << "\tReading records and building B+ Tree...done!" << endl;

			cout << "\tWriting B+ Tree to file..." << endl;
			fo << tree.write( indexDir + fileName + fileExtension ) << endl;
			cout << "\tWriting B+ Tree to file...done!" << endl;
		}
		catch( char const* p )
		{
			cout << p << endl;
			exit( 1 );
		}

		time_t end = time( NULL );
		double elapsed = (end - start) / 60.0;
		cout << "Indexing File \"" << recordDir + fileName << "\"...done!" << endl;
	}
	fo.close();
}

int main( int argc, char *argv[] )
{
	const int bucketFactor = atoi( argv[1] );
	const int order = atoi( argv[2] );
	const int ngramSize = atoi( argv[3] );
	const int indexPosition = atoi( argv[4] );
	const string unigramFile( argv[5] );
	const string recordDir( argv[6] );
	const string indexDir( argv[7] );

	cout << "Reading keys..." << endl;
	KeyMap *keyMap = new KeyMap( unigramFile );
	cout << "Reading keys...done!" << endl;

	index( ngramSize, indexPosition, bucketFactor, order, keyMap, recordDir, indexDir );

	return 0;
}

