#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include <map>

#include <climits>
#include <ctime>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>


#include "key.h"
#include "record.h"
#include "keyMap.h"
#include "bplustree.h"
#include "connection.h"
#include "minHeap.h"

using namespace std;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
class Server;
Server* server;


/*****************************************************
**************** NGRAM TREE **************************
******************************************************/
class NgramTree
{
	private:
		struct TreeInfo
		{
			int indexPosition;
			int bucketFactor;
			int order;
			Key firstKey;
			string recordFile;
			BPlusTree* tree;

			TreeInfo( int ip, int bf, int o, const string& rf, const Key& fk, BPlusTree* bpt ) : indexPosition( ip ), bucketFactor( bf ), order( o ), recordFile( rf ), firstKey( fk ), tree( bpt )
			{
			}
			TreeInfo( const TreeInfo& ti ) : bucketFactor( ti.bucketFactor ), order( ti.order ), recordFile( ti.recordFile ), firstKey( ti.firstKey ), tree( ti.tree )
			{
			}
		};

		vector<TreeInfo*> root;

		const int ngramSize;
		//lexIndexFile stands for lexicographical Index File
		const string lexIndexFile;
		const string mainIndexFile;
		const int levelsToRead;
		const KeyMap* keyMap;

		void readIndex();
		int find( const Key&, int, int ) const;
	public:
		NgramTree( int n, const string& lif, const string& mif, int ltr, const KeyMap* km ) : ngramSize( n ), lexIndexFile( lif ), mainIndexFile( mif ), levelsToRead( ltr ), keyMap( km )
		{
			readIndex();
		}
		~NgramTree();

		void retrieve( const string&, const Key&, const vector<string>&, vector<Record*>& ) const;
		int retrieveCounts( const string& query, const Key& key, const vector<string>& queryStrings, CountInteger& count, int& recordCount ) const;
};

NgramTree :: ~NgramTree()
{
	for( int i = 0; i < root.size(); i++ )
		delete root[i] -> tree;
}

void NgramTree :: readIndex()
{
        ifstream mainIfs( mainIndexFile.c_str() );
        ifstream alphabeticalIfs( lexIndexFile.c_str() );

        while( !mainIfs.eof() )
        {
		string recordFile;
                mainIfs >> recordFile;

                if( mainIfs.fail() )
                        break;

		cout << "\t\t" << "Reading indices for " << recordFile << "..." << endl;

		string nodesFile;
		mainIfs >> nodesFile;

		int indexPosition;
		int bucketFactor;
		int order;
		mainIfs >> indexPosition >> bucketFactor >> order;

		Key firstKey( ngramSize );
		mainIfs >> firstKey;

                //DiskLocation location;
		streamoff location;
                mainIfs >> location;

		Record sampleRecord( ngramSize );
		BPlusTree *bpt = new BPlusTree( sampleRecord, indexPosition, keyMap, bucketFactor, order );
		bpt -> read( nodesFile, recordFile, location, levelsToRead );

		//SEEMS USELESS NOW
		string compressedRecFile;
		string firstNgram = "";
		alphabeticalIfs >> compressedRecFile;
		for( int i = 0; i < ngramSize; i++ )
		{
			string s;
			alphabeticalIfs >> s;
			if( i != 0 )
				firstNgram += " ";
			firstNgram += s;
		}

		if( alphabeticalIfs.fail() )
			break;
		//UP TO HERE

		root.push_back( new TreeInfo( indexPosition, bucketFactor, order, recordFile, firstKey, bpt ) );

		cout << "\t\t" << "Reading indices for " << recordFile << "...Done!" << endl;
        }
}

int NgramTree :: find( const Key& key, int start, int end ) const
{
	int middle = start + ( (end - start) / 2 );

	if( start == middle || key == root[middle] -> firstKey )
		return middle;

	if( key < root[middle] -> firstKey )
		return find( key, start, middle );
	else
		return find( key, middle, end );
}

void NgramTree :: retrieve( const string& query, const Key& key, const vector<string>& queryStrings, vector<Record*>& results ) const
{
	int index = find( key, 0, root.size() );

	DiskLocation location = root[index] -> tree -> query( key );
	ifstream fo( ( root[index] -> recordFile ).c_str() );
	fo.seekg( location );
	int recordCount = root[index] -> bucketFactor;
	bool matchFlag = false;

	do
	{
		while( !fo.eof() && recordCount > 0 )
		{
			Record record( 0, queryStrings );
			try
			{
				fo >> record;
			}
			catch( const char* cp )
			{
				break;
			}

			if( record.match( queryStrings ) )
			{
				matchFlag = true;
				results.push_back( new Record( record ) );
				break;
			}
			recordCount--;
		}
	
		if( recordCount > 0 && !matchFlag )
		{
			fo.clear();
			fo.close();
			index++;
			if( index >= root.size() )
				return;
			fo.open( ( root[index] -> recordFile ).c_str() );
		}
		else
			break;
	}
	while( true );

	if( matchFlag )
	{
		while( 1 )
		{
			do
			{
				Record record( 0, queryStrings );
				try
				{
					fo >> record;
				}
				catch( const char* cp )
				{
					break;
				}

				if( record.match( queryStrings ) )
					results.push_back( new Record( record ) );
				else
				{
					matchFlag = false;
					break;
				}
			}
			while( 1 );

			if( !matchFlag )
				break;

			index++;
			if( index == root.size() )
				break;
			fo.clear();
			fo.close();
			fo.open( ( root[index] -> recordFile ).c_str() );
		}
	}
	fo.close();
}

int NgramTree :: retrieveCounts( const string& query, const Key& key, const vector<string>& queryStrings, CountInteger& count, int& recCount ) const
{
	int index = find( key, 0, root.size() );

	DiskLocation location = root[index] -> tree -> query( key );
	ifstream fo( ( root[index] -> recordFile ).c_str() );
	fo.seekg( location );
	int recordCount = root[index] -> bucketFactor;
	bool matchFlag = false;
	map<string, bool> uniqueFollowingWords;

	do
	{
		while( !fo.eof() && recordCount > 0 )
		{
			Record record( 0, queryStrings );
			try
			{
				fo >> record;
			}
			catch( const char* cp )
			{
				break;
			}

			if( record.match( queryStrings ) )
			{
				matchFlag = true;
				count += record.ngramCount;
				recCount++;
				uniqueFollowingWords[record.words[queryStrings.size() - 1]] = true;
				break;
			}
			recordCount--;
		}
	
		if( recordCount > 0 && !matchFlag )
		{
			fo.clear();
			fo.close();
			index++;
			if( index >= root.size() )
				return uniqueFollowingWords.size();
			fo.open( ( root[index] -> recordFile ).c_str() );
		}
		else
			break;
	}
	while( true );

	if( matchFlag )
	{
		while( 1 )
		{
			do
			{
				Record record( 0, queryStrings );
				try
				{
					fo >> record;
				}
				catch( const char* cp )
				{
					break;
				}

				if( record.match( queryStrings ) )
				{
					count += record.ngramCount;
					recCount++;
					if( uniqueFollowingWords.find( record.words[queryStrings.size()-1] ) == uniqueFollowingWords.end() )
						uniqueFollowingWords[record.words[queryStrings.size()-1]] = true;
				}
				else
				{
					matchFlag = false;
					break;
				}
			}
			while( 1 );

			if( !matchFlag )
				break;

			index++;
			if( index == root.size() )
				break;
			fo.clear();
			fo.close();
			fo.open( ( root[index] -> recordFile ).c_str() );
		}
	}
	fo.close();

	return uniqueFollowingWords.size();
}



#define MAX_CACHE_SIZE	10000

struct CacheRecord
{
	string ngram;
	long timeStamp;

	CacheRecord() : ngram( "" ), timeStamp( -1 )
	{}
	CacheRecord( const string& s, long t ) : ngram( s ), timeStamp( t )
	{}

	bool operator<( const CacheRecord& cr )
	{
		return timeStamp < cr.timeStamp;
	}
};

struct CacheCount
{
	CountInteger count;
	int recordCount;
	int uniqueFollowingWords;

	CacheCount() : count( 0 ), recordCount( 0 ), uniqueFollowingWords( 0 )
	{}
	CacheCount( const CountInteger& c, int r, int u ) : count( c ), recordCount( r ), uniqueFollowingWords( u )
	{}
};

/*****************************************************
**************** MAIN TREE ***************************
******************************************************/
class MainTree
{
	private:
		const int minNgram;
		const int maxNgram;

		const string mainRecordDir;
		const string unigramFile;
		const string indexFilesMainDir;
		const string mainIndexFileName;

		const KeyMap* keyMap;
		vector<NgramTree*> allTrees;

		long timeStamp;
		map<string, CacheCount> cache;
		MinHeap<CacheRecord> cacheHistory;

		void initialize();

		bool isCached( const string& );
		int retrieveFromCache( const string&, CountInteger&, int& );
		void insertIntoCache( const string&, CountInteger&, int, int );

	public:
		MainTree( int min, int max, const string& mrd, const string& ifmd, const string& mifn, const KeyMap* km ) : minNgram( min ), maxNgram( max ), mainRecordDir( mrd ), indexFilesMainDir( ifmd ), mainIndexFileName( mifn ), keyMap( km ), timeStamp( 0 )
		{
			initialize();
		}
		~MainTree();

		void query( const string&, int, const vector<string>&, vector<Record*>& ) const;
		int countQuery( const string&, int, const vector<string>&, CountInteger&, int& );
};

MainTree :: ~MainTree()
{
	for( int i = minNgram; i <= maxNgram; i++ )
		delete allTrees[i];
	delete keyMap;
}

void MainTree :: initialize()
{
	//This can actually vary for each B+Tree
	const int levelsToRead = 2;

	//Unigrams are kept in memory;	
	//To make array indices equal to ngramSize
	for( int i = 0; i < minNgram; i++ )
		allTrees.push_back( NULL );

	for( int ngramSize = minNgram; ngramSize <= maxNgram; ngramSize++ )
	{
		cout << "\t" << "Reading " << ngramSize << "-gram indices..." << endl;
		const Record sampleRecord( ngramSize );

		ostringstream oss;
		oss << ngramSize;
		const string mainIndexFile = indexFilesMainDir + oss.str() + "gms/" + mainIndexFileName;
		const string lexIndexFile = mainRecordDir + oss.str() + "gms/" + oss.str() + "gm.idx";

		allTrees.push_back( new NgramTree( ngramSize, lexIndexFile, mainIndexFile, levelsToRead, keyMap ) );
		cout << "\t" << "Reading " << ngramSize << "-gram indices...Done!" << endl << endl;
	}
}

bool MainTree :: isCached( const string& qString )
{
	map<string, CacheCount> :: iterator it = cache.find( qString );
	return ( it != cache.end() );
}

int MainTree :: retrieveFromCache( const string& qString, CountInteger& count, int& recordCount )
{
	map<string, CacheCount> :: iterator it = cache.find( qString );
	if( it != cache.end() )
	{
		count = ( it -> second ).count;
		recordCount = ( it -> second ).recordCount;
		return ( it -> second ).uniqueFollowingWords;
	}
	return 0;
}

void MainTree :: insertIntoCache( const string& qString, CountInteger& count, int recordCount, int uniqueFollowingWords )
{
	map<string, CacheCount> :: iterator it = cache.find( qString );
	if( it != cache.end() )
		return;

	if( timeStamp == LONG_MAX )
	{
		//Restart everything when timeStamp cannot be increased anymore as it is type long
		//This should definitely change in future
		timeStamp = 0;
		cache.clear();
		while( !cacheHistory.isEmpty() )
		{
			cacheHistory.pop();
		}
	}

	CacheRecord r;
	if( cache.size() == MAX_CACHE_SIZE )
	{
		r = cacheHistory.top();
		cacheHistory.pop();
		it = cache.find( r.ngram );
		cache.erase( it );
	}
	timeStamp++;
	CacheRecord newRecord( qString, timeStamp );
	CacheCount newCount( count, recordCount, uniqueFollowingWords );
	cacheHistory.push( newRecord );
	cache.insert( pair<string, CacheCount>( qString, newCount ) );
}

void MainTree :: query( const string& qString, int indexPosition, const vector<string>& queryStrings, vector<Record*>& results ) const
{
	const int ngramSize = queryStrings.size();
	if( ngramSize > maxNgram || ngramSize <= 0 )
		throw( "Invalid N-gram!" );

	if( ngramSize == 1 )
	{
		try
		{
			CountInteger ngramCount = keyMap -> getCount( qString );
			results.push_back( new Record( ngramCount, queryStrings ) );
		}
		catch( const char* cp )
		{
			throw cp;
		}
	}
	else
	{
		try
		{
			Key key;
			keyMap -> getKey( queryStrings, indexPosition, key );

			allTrees[ngramSize] -> retrieve( qString, key, queryStrings, results );
		}
		catch( const char* cp )
		{
			throw( cp );
		}
	}
}

int MainTree :: countQuery( const string& qString, int indexPosition, const vector<string>& queryStrings, CountInteger& count, int& recordCount )
{
	const int ngramSize = queryStrings.size();
	if( ngramSize > maxNgram || ngramSize <= 0 )
		throw( "Invalid N-gram!" );

	if( ngramSize == 1 )
	{
		try
		{
			count = keyMap -> getCount( qString );
		}
		catch( const char* cp )
		{
			return 0;
			//throw( cp );
		}
	}
	else
	{
		try
		{
			if( isCached( qString ) )
			{
				return retrieveFromCache( qString, count, recordCount );
			}

			Key key;
			keyMap -> getKey( queryStrings, indexPosition, key );

			int uniqueFollowingWords = allTrees[ngramSize] -> retrieveCounts( qString, key, queryStrings, count, recordCount );
			insertIntoCache( qString, count, recordCount, uniqueFollowingWords );
			return uniqueFollowingWords;
		}
		catch( const char* cp )
		{
			return 0;
			//throw( cp );
		}
	}
}


class Smoothing
{
	private:
		//These should be const as well but this is not possible on 32 bit architectures
		CountInteger totalCounts[6];
		long uniqueCounts[6];

		//In principle this should be const too
		double dAbs[6];

		double dKNHigher[6];
		double dKNLower[5];

		double dMCHigher[6];
		double dMCLower[5];
		double deltaMC;

		double kDir[6];

		double kDKNHigher[6];
		double kDKNLower[5];
		double deltaDKN;

		void getNextQueries( const string&, const vector<string>&, string&, vector<string>&, string&, vector<string>& );

		void absDiscountHelper( const string&, const vector<string>&, const string&, const vector<string>&, MainTree*, double, double&, double& );
		double kNDiscountHelper( const string&, const vector<string>&, MainTree* );
		double missingCountsHelper( const string&, const vector<string>&, MainTree*, MainTree* );
		void dirichletHelper( const string&, const vector<string>&, const string&, const vector<string>&, MainTree*, double, double&, double& );
		double dKNHelper( const string&, const vector<string>&, MainTree*, MainTree* );

	public:
		Smoothing();

		double getProb( const string&, const vector<string>&, MainTree* );
		double getAbsDiscountProb( const string&, const vector<string>&, MainTree* );
		double getKNDiscountProb( const string&, const vector<string>&, MainTree*, MainTree* );
		double getMissingCountsProb( const string&, const vector<string>&, MainTree*, MainTree* );
		double getDirichletProb( const string&, const vector<string>&, MainTree* );
		double getDKNProb( const string&, const vector<string>&, MainTree*, MainTree* );
};

Smoothing :: Smoothing()
{
	totalCounts[0] = 0;

	totalCounts[1] = 7975133;
	totalCounts[1] *= 128513;

	totalCounts[2] = 264403;
	totalCounts[2] *= 3445061;

	totalCounts[3] = 1094824961;
	totalCounts[3] *= 675;
	totalCounts[3] -= 1;
	//totalCounts[3] = 2820636827;
	//totalCounts[3] *= 262;

	//totalCounts[4] = 22018152571;
	//totalCounts[4] *= 24;
	totalCounts[4] = 351120041;
	totalCounts[4] *= 1505;
	totalCounts[4] -= 1;

	totalCounts[5] = 205926421;
	totalCounts[5] *= 1789;

	uniqueCounts[0] = 0;
	uniqueCounts[1] = 13588391;
	uniqueCounts[2] = 314843401;
	uniqueCounts[3] = 977069902;
	uniqueCounts[4] = 1333820466;
	uniqueCounts[5] = 1214460675;

	dAbs[0] = 0, dAbs[1] = 0, dAbs[2] = 0.94914893, dAbs[3] = 0.99209517, dAbs[4] = 0.99951819, dAbs[5] = 0.99995625;

	dKNHigher[0] = 0, dKNHigher[1] = 0, dKNHigher[2] = 0.98520067, dKNHigher[3] = 0.99577209, dKNHigher[4] = 0.99933982, dKNHigher[5] = 0.99996885;
	dKNLower[0] = 0, dKNLower[1] = 0, dKNLower[2] = 0.80424828, dKNLower[3] = 0.91786366, dKNLower[4] = 0.98042218;

	dMCHigher[0] = 0, dMCHigher[1] = 0, dMCHigher[2] = 0.9851726, dMCHigher[3] = 0.99581479, dMCHigher[4] = 0.99931418, dMCHigher[5] = 0.99996568;
	dMCLower[0] = 0, dMCLower[1] = 0, dMCLower[2] = 0.97914173, dMCLower[3] = 0.99992424, dMCLower[4] = 0.99985808;
	deltaMC = (0.095564851 + 1) / (0.095564851 + 40);

	kDir[0] = 0, kDir[1] = 0, kDir[2] = 7.8440119, kDir[3] = 6.6305746, kDir[4] = 6.9277921, kDir[5] = 7.3675802;

	kDKNHigher[0] = 0, kDKNHigher[1] = 0, kDKNHigher[2] = 3.192501, kDKNHigher[3] = 2.8096546125, kDKNHigher[4] = 3.23767800788, kDKNHigher[5] = 3.58515159429;
	kDKNLower[0] = 0, kDKNLower[1] = 0, kDKNLower[2] = 0.156310135, kDKNLower[3] = 0.205351523914, kDKNLower[4] = 0.238121750327;
	deltaDKN = 0.119330390707;

	const int MIN_NGRAM = 40;
	for( int i = 2; i < 6; i++ )
	{
		dAbs[i] *= MIN_NGRAM;
		dKNHigher[i] *= MIN_NGRAM;
		dMCHigher[i] *= MIN_NGRAM;
	}
}

void Smoothing :: getNextQueries( const string& qString, const vector<string>& queryWords, string& qString1, vector<string>& q1Words, string& qString2, vector<string>& q2Words )
{
	for( int i = 0; i < queryWords.size(); i++ )
	{
		if( i != queryWords.size() - 1 )
		{
			qString1 += queryWords[i] + " ";
			q1Words.push_back( queryWords[i] );
		}
		if( i != 0 )
		{
			qString2 += queryWords[i] + " ";
			q2Words.push_back( queryWords[i] );
		}
	}
	qString1 += "_";
	q1Words.push_back( "_" );
	qString2.resize( qString2.length() - 1 );
}

double Smoothing :: getProb( const string& qString, const vector<string>& queryWords, MainTree* tree )
{
	CountInteger c = 0;
	int n = 0;
	tree -> countQuery( qString, 0, queryWords, c, n );

	CountInteger c2 = 0;
	if( queryWords.size() == 1 )
		c2 = totalCounts[1];
	else
	{
		string qString2 = "";
		vector<string> q2Words;
		for( int i = 0; i < ( queryWords.size() - 1 ); i++ )
		{
			qString2 += queryWords[i] + " ";
			q2Words.push_back( queryWords[i] );
		}
		qString2.resize( qString2.length() - 1 );

		int n2 = 0;
		tree -> countQuery( qString2, 0, q2Words, c2, n2 );
	}

	if( c2 == 0 )
		return 0;

	return ( (double) c ) / c2;
}

double Smoothing :: getAbsDiscountProb( const string& qString, const vector<string>& queryWords, MainTree* tree )
{
	if( queryWords.size() == 1 )
	{
		//If the word is not in the vocabulary, then it will get a 0 probability (<UNK> is not used)
		//It's also possible to include a ground case with a 0 order model (1 / (size_of_vocabulary))
		//Or we can assign an arbitrary count (100 in Dr. Yuret's case) to unseen words though this wouldn't be probabilistically sound
		CountInteger c = 0;
		int n = 0;
		tree -> countQuery( qString, 0, queryWords, c, n );
		return ( (double) c ) / totalCounts[1];
	}

	string qString1 = "", qString2 = "";
	vector<string> q1Words, q2Words;
	getNextQueries( qString, queryWords, qString1, q1Words, qString2, q2Words );

	double alpha = 0, gamma = 0;
	double dConstant = dAbs[queryWords.size()];
	absDiscountHelper( qString, queryWords, qString1, q1Words, tree, dConstant, alpha, gamma );

	return ( alpha + ( gamma * getAbsDiscountProb( qString2, q2Words, tree ) ) );
}

void Smoothing :: absDiscountHelper( const string& qString, const vector<string>& queryWords, const string& qString1, const vector<string>& q1Words, MainTree* tree, double dConstant, double& alpha, double& gamma )
{
	CountInteger count = 0;
	int n = 0;
	tree -> countQuery( qString, 0, queryWords, count, n );

	CountInteger count1 = 0;
	int n1 = 0;
	tree -> countQuery( qString1, 0, q1Words, count1, n1 );

	alpha = ( count - dConstant );
	if( alpha < 0 || count1 == 0 )
		alpha = 0;
	else
		alpha /= count1;

	if( count1 != 0 )
		gamma = ( n1 * dConstant ) / count1;

	//If the context does not exist, then use the lower order model directly
	if( gamma == 0 )
		gamma = 1;
}

double Smoothing :: getKNDiscountProb( const string& qString, const vector<string>& queryWords, MainTree* tree0, MainTree* tree1 )
{
	if( queryWords.size() == 1 )
	{
		CountInteger c = 0;
		int n = 0;
		tree0 -> countQuery( qString, 0, queryWords, c, n );
		return ( (double) c ) / totalCounts[1];
	}

	string qString1 = "", qString2 = "";
	vector<string> q1Words, q2Words;
	getNextQueries( qString, queryWords, qString1, q1Words, qString2, q2Words );

	double alpha = 0, gamma = 0;
	double dConstant = dKNHigher[queryWords.size()];
	absDiscountHelper( qString, queryWords, qString1, q1Words, tree0, dConstant, alpha, gamma );

	return ( alpha + ( gamma * kNDiscountHelper( qString2, q2Words, tree1 ) ) );
}

double Smoothing :: kNDiscountHelper( const string& qString, const vector<string>& queryWords, MainTree* tree )
{
	if( queryWords.size() == 1 )
	{
		string qs = "_ " + qString;
		vector<string> qWords;
		qWords.push_back( "_" );
		qWords.push_back( qString );
		CountInteger c = 0;
		int n = 0;
		tree -> countQuery( qs, 1, qWords, c, n );

		return ( (double) n ) / uniqueCounts[2];
	}

	string qString2 = "_ " + qString;
	vector<string> q2Words;
	q2Words.push_back( "_" );
	for( int i = 0; i < queryWords.size(); i++ )
		q2Words.push_back( queryWords[i] );

	CountInteger count2 = 0;
	int n2 = 0;
	tree -> countQuery( qString2, 1, q2Words, count2, n2 );

	string qString3 = "_ ", qString4 = "";
	vector<string> q3Words, q4Words;
	q3Words.push_back( "_" );
	getNextQueries( qString, queryWords, qString3, q3Words, qString4, q4Words );

	CountInteger count3 = 0;
	int n3 = 0;
	int r = tree -> countQuery( qString3, 1, q3Words, count3, n3 );

	double alpha = ( n2 - dKNLower[queryWords.size()] );
	if( alpha < 0 || n3 == 0 )
		alpha = 0;
	else
		alpha /= n3;
	double gamma = 0;
	if( n3 != 0 )
		gamma = ( r * dKNLower[queryWords.size()] ) / n3;

	//If the context does not exist, then use the lower order model directly
	if( gamma == 0 )
		gamma = 1;

	return ( alpha + ( gamma * kNDiscountHelper( qString4, q4Words, tree ) ) );
}

double Smoothing :: getMissingCountsProb( const string& qString, const vector<string>& queryWords, MainTree* tree0, MainTree* tree1 )
{
	if( queryWords.size() == 1 )
	{
		CountInteger c = 0;
		int n = 0;
		tree0 -> countQuery( qString, 0, queryWords, c, n );
		return ( (double) c ) / totalCounts[1];
	}

	string qString1 = "", qString2 = "";
	vector<string> q1Words, q2Words;
	getNextQueries( qString, queryWords, qString1, q1Words, qString2, q2Words );

	double alpha = 0, gamma = 0;
	double dConstant = dMCHigher[queryWords.size()];
	absDiscountHelper( qString, queryWords, qString1, q1Words, tree0, dConstant, alpha, gamma );

	return ( alpha + ( gamma * missingCountsHelper( qString2, q2Words, tree0, tree1 ) ) );
}

double Smoothing :: missingCountsHelper( const string& qString, const vector<string>& queryWords, MainTree* tree0, MainTree* tree1 )
{
	if( queryWords.size() == 1 )
	{
		string qs = "_ " + qString;
		vector<string> qWords;
		qWords.push_back( "_" );
		qWords.push_back( qString );
		CountInteger c = 0;
		int nn = 0;
		tree1 -> countQuery( qs, 1, qWords, c, nn );

		return ( (double) nn ) / uniqueCounts[2];
	}

	CountInteger count = 0;
	int n = 0;
	tree0 -> countQuery( qString, 0, queryWords, count, n );

	string qString2 = "_ " + qString;
	vector<string> q2Words;
	q2Words.push_back( "_" );
	for( int i = 0; i < queryWords.size(); i++ )
		q2Words.push_back( queryWords[i] );

	CountInteger count2 = 0;
	int n2 = 0;
	tree1 -> countQuery( qString2, 1, q2Words, count2, n2 );

	int modN = n2 + ( deltaMC * ( count - count2 ) );

	string qString3 = "_ ", qString4 = "";
	vector<string> q3Words, q4Words;
	q3Words.push_back( "_" );
	getNextQueries( qString, queryWords, qString3, q3Words, qString4, q4Words );

	CountInteger count3 = 0;
	int n3 = 0;
	int r = tree1 -> countQuery( qString3, 1, q3Words, count3, n3 );

	string qString5 = "";
	vector<string> q5Words;
	for( int i = 0; i < (queryWords.size() - 1); i++ )
	{
		qString5 += queryWords[i] + " ";
		q5Words.push_back( queryWords[i] );
	}
	qString5 += "_";
	q5Words.push_back( "_" );

	CountInteger count5 = 0;
	int n5 = 0;
	tree0 -> countQuery( qString5, 0, q5Words, count5, n5 );
	int modN2 = n3 + ( deltaMC * ( count5 - count3 ) );

	double alpha = modN - dMCLower[queryWords.size()];
	if( alpha < 0 || modN2 == 0 )
		alpha = 0;
	else
		alpha /= modN2;

	double gamma = 0;
	if( modN2 != 0 )
		gamma = ( n5 * dMCLower[queryWords.size()] ) / modN2;

	//If the context does not exist, then use the lower order model directly
	if( gamma == 0 )
		gamma = 1;

	return ( alpha + ( gamma * missingCountsHelper( qString4, q4Words, tree0, tree1 ) ) );
}

double Smoothing :: getDirichletProb( const string& qString, const vector<string>& queryWords, MainTree* tree )
{
	if( queryWords.size() == 1 )
	{
		CountInteger c = 0;
		int n = 0;
		tree -> countQuery( qString, 0, queryWords, c, n );
		return ( (double) c ) / totalCounts[1];
	}

	string qString1 = "", qString2 = "";
	vector<string> q1Words, q2Words;
	getNextQueries( qString, queryWords, qString1, q1Words, qString2, q2Words );

	double alpha = 0, gamma = 0;
	double kConstant = kDir[queryWords.size()];
	dirichletHelper( qString, queryWords, qString1, q1Words, tree, kConstant, alpha, gamma );

	return ( alpha + ( gamma * getDirichletProb( qString2, q2Words, tree ) ) );
}

void Smoothing :: dirichletHelper( const string& qString, const vector<string>& queryWords, const string& qString1, const vector<string>& q1Words, MainTree* tree, double kConstant, double& alpha, double& gamma )
{
	CountInteger count = 0;
	int n = 0;
	tree -> countQuery( qString, 0, queryWords, count, n );

	CountInteger count1 = 0;
	int n1 = 0;
	tree -> countQuery( qString1, 0, q1Words, count1, n1 );

	string qString2 = "";
	vector<string> q2Words;
	for( int i = 0; ( i < queryWords.size() - 1 ); i++ )
	{
		qString2 += queryWords[i];
		q2Words.push_back( queryWords[i] );
		if( i != ( queryWords.size() - 2 ) )
			qString2 += " ";
	}

	CountInteger count2 = 0;
	int n2 = 0;
	tree -> countQuery( qString2, 0, q2Words, count2, n2 );

	double a = kConstant * ( count2 - count1 );
	if( a < 1.0 )
		a = 1.0;

	alpha = ( (double) count ) / ( count1 + a );
	gamma = a / ( count1 + a );

	//If the context does not exist, then use the lower order model directly
	if( gamma == 0 )
		gamma = 1;
}

double Smoothing :: getDKNProb( const string& qString, const vector<string>& queryWords, MainTree* tree0, MainTree* tree1 )
{
	if( queryWords.size() == 1 )
	{
		CountInteger c = 0;
		int n = 0;
		tree0 -> countQuery( qString, 0, queryWords, c, n );
		return ( (double) c ) / totalCounts[1];
	}

	string qString1 = "", qString2 = "";
	vector<string> q1Words, q2Words;
	getNextQueries( qString, queryWords, qString1, q1Words, qString2, q2Words );

	double alpha = 0, gamma = 0;
	double kConstant = kDKNHigher[queryWords.size()];
	dirichletHelper( qString, queryWords, qString1, q1Words, tree0, kConstant, alpha, gamma );

	return ( alpha + ( gamma * dKNHelper( qString2, q2Words, tree0, tree1 ) ) );
}

double Smoothing :: dKNHelper( const string& qString, const vector<string>& queryWords, MainTree* tree0, MainTree* tree1 )
{
	if( queryWords.size() == 1 )
	{
		string qs = "_ " + qString;
		vector<string> qWords;
		qWords.push_back( "_" );
		qWords.push_back( qString );
		CountInteger c = 0;
		int nn = 0;
		tree1 -> countQuery( qs, 1, qWords, c, nn );

		return ( (double) nn ) / uniqueCounts[2];
	}

	CountInteger count = 0;
	int n = 0;
	tree0 -> countQuery( qString, 0, queryWords, count, n );

	string qString2 = "_ " + qString;
	vector<string> q2Words;
	q2Words.push_back( "_" );
	for( int i = 0; i < queryWords.size(); i++ )
		q2Words.push_back( queryWords[i] );

	CountInteger count2 = 0;
	int n2 = 0;
	tree1 -> countQuery( qString2, 1, q2Words, count2, n2 );

	int modN = n2 + ( deltaDKN * ( count - count2 ) );

	string qString3 = "_ ", qString4 = "";
	vector<string> q3Words, q4Words;
	q3Words.push_back( "_" );
	getNextQueries( qString, queryWords, qString3, q3Words, qString4, q4Words );

	CountInteger count3 = 0;
	int n3 = 0;
	int r = tree1 -> countQuery( qString3, 1, q3Words, count3, n3 );

	string qString5 = "", qString6 = "";
	vector<string> q5Words, q6Words;
	for( int i = 0; i < (queryWords.size() - 1); i++ )
	{
		qString5 += queryWords[i] + " ";
		q5Words.push_back( queryWords[i] );

		qString6 += queryWords[i] + " ";
		q6Words.push_back( queryWords[i] );
	}
	qString6 += "_";
	q6Words.push_back( "_" );

	CountInteger count5 = 0;
	int n5 = 0;
	tree0 -> countQuery( qString5, 0, q5Words, count5, n5 );

	CountInteger count6 = 0;
	int n6 = 0;
	tree0 -> countQuery( qString6, 0, q6Words, count6, n6 );

	double a = kDKNLower[queryWords.size()] * ( count5 - count6 );
	if( a < 1 )
		a = 1;

	int modN2 = n3 + ( deltaDKN * ( count6 - count3 ) );

	double alpha = modN / ( modN2 + a );
	double gamma = a / ( modN2 + a );

	return ( alpha + ( gamma * dKNHelper( qString4, q4Words, tree0, tree1 ) ) );	
}

/*****************************************************
**************** SERVER ******************************
******************************************************/
#define MAX_LINE_LENGTH 256
#define MIN_NGRAM	1
#define MAX_NGRAM	5

#define NO_SMOOTHING_COUNTS_ONLY	0
#define NO_SMOOTHING_NGRAMS	1
#define NO_SMOOTHING_PROB	2
#define ABSOLUTE_DISCOUNTING	3
#define KN_DISCOUNTING	4
#define KN_MC_DISCOUNTING	5
#define DIRICHLET_DISCOUNTING	6
#define DKN_DISCOUNTING	7


class Server : public Connection
{
	private:
		const int maxNumOfConnections;
		const int backLog;
		map<int, pthread_t> connections;

		map<string, vector<MainTree*> > trees;

		Smoothing smoothing;

		void bindSocket();
		void listenPort();
		void acceptConnection();
		void closeConnection( int );

		int analyzeQueryString( const char*, int, string&, string&, vector<string>& ) const;
		int interface( string&, string&, vector<string>& ) const;

		void query( const char*, int, string&, vector<string>&, vector<Record*>& );
		int countQuery( const char*, int, string&, vector<string>&, CountInteger&, int& );
		static void encodeString( const string&, byte* );
		static void* handleConnection( void* );

	public:
		Server( const map<string, vector<MainTree*> >& ts, const int, const int, const int );

		void runInteractive();
		void run();
};

Server :: Server( const map<string, vector<MainTree*> >& m, const int port, const int mc, const int bl ) : Connection( INADDR_ANY, port ), maxNumOfConnections( mc ), backLog( bl )
{
	map<string, vector<MainTree*> > :: const_iterator it;
	for( it = m.begin(); it != m.end(); it++ )
	{
		string language = it -> first;

		const vector<MainTree*>& ts = it -> second;
		for( int i = 0; i < ts.size(); i++ )
			trees[language].push_back( ts[i] );
	}

	try
	{
		bindSocket();
		listenPort();
	}
	catch( const char* cp )
	{
		perror( cp );
	}
}

void Server :: bindSocket()
{
	int bindResponse = bind( socketFileDesc, ( SocketAddress* ) &address, sizeof( SocketAddress ) );

	if( bindResponse == -1 )
		throw( "bind" );

}

void Server :: listenPort()
{
	int listenResponse = listen( socketFileDesc, backLog );

	if( listenResponse == -1 )
		throw( "listen" );

}

void Server :: acceptConnection()
{
	SocketAddressIn incomingAddress;
	int addressSize = sizeof( SocketAddressIn );
	int newSocket = accept( socketFileDesc, (SocketAddress *) &incomingAddress, (socklen_t *) &addressSize );

	if( connections.size() == maxNumOfConnections )
	{
		close( newSocket );
		return;
	}

	if( newSocket == -1 )
		throw( "accept" );

	cout << endl << "Accepted a new connection from " << inet_ntoa( incomingAddress.sin_addr ) << endl;
	cout << "Current number of connections: " << connections.size() + 1 << endl;

	connections[newSocket] = 0;
	map<int, pthread_t> :: iterator it = connections.find( newSocket );
	pthread_create( &(it -> second), NULL, handleConnection, (void*) &(it -> first) );
}

void Server :: closeConnection( int socket )
{
	cout << "Closing connection ID " << connections.find( socket ) -> second << endl;
	connections.erase( socket );
	close( socket );
}

int Server :: analyzeQueryString( const char* line, int lineLength, string& qString, string& language, vector<string>& words ) const
{
	words.clear();
	for( int i = 0; i < lineLength; i++ )
	{
		string word = "";
		while( i < lineLength && line[i] != ' ' && line[i] != '\0' )
		{
			word += line[i];
			i++;
		}

		words.push_back( word );
	}

	language = words[0];
	if( trees.find( language ) == trees.end() )
		throw( "Invalid Language!" );

	words.erase( words.begin() );
	if( words.size() < MIN_NGRAM || words.size() > MAX_NGRAM )
		throw( "Invalid N-gram!" );

	qString = "";
	for( int i = 0; i < words.size(); i++ )
	{
		qString += words[i];
		if( i != ( words.size() - 1 ) )
			qString += " ";
	}

	int specialSymbolPos = 0;
	int flagCounter = 0;
	for( int i = 0; i < words.size(); i++ )
	{
		if( i == 0 )
		{
			flagCounter++;
		}
		else
		{
			if( words[i] == SPECIAL_SYMBOL )
			{
				if( words[i-1] != SPECIAL_SYMBOL )
					flagCounter++;
			}
			else
			{
				if( words[i-1] == SPECIAL_SYMBOL )
				{
					flagCounter++;
					specialSymbolPos = i;
				}
			}
		}
	}

	if( flagCounter >= 4 )
		throw( "Invalid N-gram!" );

	return specialSymbolPos;
}

int Server :: interface( string& qString, string& language, vector<string>& words ) const
{
	cout << ">> ";
	char line[MAX_LINE_LENGTH];
	cin.getline( line, MAX_LINE_LENGTH );
	if( cin.fail() )
		throw( "Invalid Query!" );

	qString = "";
	return analyzeQueryString( line, cin.gcount(), qString, language, words );
}

void Server :: runInteractive()
{
	while( 1 )
	{
		try
		{
			string qString;
			string language;
			vector<string> queryWords;
			int treeNo = interface( qString, language, queryWords );

			if( qString == "!exit" )
				break;

			CountInteger count = 0;
			int n = 0;

			map<string, vector<MainTree*> > :: const_iterator it = trees.find( language );
			const vector<MainTree*>& ts = it -> second;
			ts[treeNo] -> countQuery( qString, 0, queryWords, count, n );

			cout << "CNT: " << count << endl;
			//cout << "PRB: " << smoothing.getProb( qString, queryWords, trees[0] ) << endl;
			//cout << "ABS: " << smoothing.getAbsDiscountProb( qString, queryWords, trees[0] ) << endl;
			//cout << "KN: " << smoothing.getKNDiscountProb( qString, queryWords, trees[0], trees[1] ) << endl;
			//cout << "MC: " << smoothing.getMissingCountsProb( qString, queryWords, trees[0], trees[1] ) << endl;
			//cout << "DIR: " << smoothing.getDirichletProb( qString, queryWords, trees[0] ) << endl;
			//cout << "DKN: " << smoothing.getDKNProb( qString, queryWords, trees[0], trees[1] ) << endl;

			//vector<Record*> results;
			//trees[treeNo] -> query( qString, treeNo, queryWords, results );
			//for( int i = 0; i < results.size(); i++ )
			//{
			//	cout << *results[i] << endl;
			//	delete results[i];
			//}
		}
		catch( const char* cp )
		{
			cerr << cp << endl;
			cin.clear();
		}
	}
}

void Server :: run()
{
	while( 1 )
	{
		try
		{
			acceptConnection();
		}
		catch( const char* cp )
		{
			perror( cp );
			break;
		}
	}
}

//This function is only needed by the static handleConnection() function
void Server :: query( const char* stringBuffer, int bufferLength, string& qString, vector<string>& words, vector<Record*>& results )
{
	string language;
	int treeNo = analyzeQueryString( stringBuffer, bufferLength, qString, language, words );
	map<string, vector<MainTree*> > :: const_iterator it = trees.find( language );
	const vector<MainTree*>& ts = it -> second;
	ts[treeNo] -> query( qString, treeNo, words, results );
}

//This function is only needed by the static handleConnection() function
int Server :: countQuery( const char* stringBuffer, int bufferLength, string& qString, vector<string>& words, CountInteger& count, int& n )
{
	string language;
	int treeNo = analyzeQueryString( stringBuffer, bufferLength, qString, language, words );
	map<string, vector<MainTree*> > :: const_iterator it = trees.find( language );
	const vector<MainTree*>& ts = it -> second;
	return ts[treeNo] -> countQuery( qString, treeNo, words, count, n );
}

void Server :: encodeString( const string& str, byte* byteArray )
{
	for( int i = 0; i < str.length(); i++ )
		byteArray[i] = str[i];
}

void* Server :: handleConnection( void* p )
{
	int socket = *((int*) p);

	while( 1 )
	{
		string qString;
		string language;
		vector<string> queryWords;
		vector<Record*> results;
		int method = 0;
		int messageSize = 0;
		byte* message = NULL;

		try
		{
			//Receive the type of method (i.e. no smoothing, abs discounting, etc.)
			const int bufferSize = 5;
			byte buffer[bufferSize];
			buffer[1] = '\0';
			server -> receive( buffer, 1, socket );
			method = atoi( (const char*) buffer );

			//Receive the incoming string length
			buffer[bufferSize-1] = '\0';
			server -> receive( buffer, bufferSize-1, socket );
			const int length = atoi( (const char*) buffer );

			//Receive the incoming string now
			byte stringBuffer[length];
			server -> receive( stringBuffer, length, socket );
			//qString = string( (const char*) stringBuffer, length );

			pthread_mutex_lock( &mutex );

			CountInteger count = 0;
			int n = 0, r = 0;
			double probability = 0.0;
			int startPointer = 0;
			ostringstream oss;
			map<string, vector<MainTree*> > :: iterator it;
			vector<MainTree*> ts;

			switch( method )
			{
				case NO_SMOOTHING_COUNTS_ONLY:
					server -> countQuery( (const char*) stringBuffer, length, qString, queryWords, count, n );

					oss << count;
					messageSize = oss.str().length() + 1;
					message = new byte[messageSize];
					encodeString( oss.str(), message );
					message[messageSize-1] = '\n';
	
					break;
	
				case NO_SMOOTHING_NGRAMS:
					server -> query( (const char*) stringBuffer, length, qString, queryWords, results );

					oss << results.size();
					messageSize += oss.str().length() + 1;
					oss.str( "" );
					for( int i = 0; i < results.size(); i++ )
					{
						for( int j = 0; j < results[i] -> numOfWords; j++ )
						{
							messageSize += (results[i] -> words)[j].length();
						}
						messageSize += results[i] -> numOfWords;
						oss << results[i] -> ngramCount;
						messageSize += oss.str().length() + 1;
						oss.str( "" );
					}
	
					message = new byte[messageSize];
					oss << results.size();
					encodeString( oss.str(), message );
					message[oss.str().length()] = '\n';
					startPointer += oss.str().length() + 1;
					oss.str( "" );
					for( int i = 0; i < results.size(); i++ )
					{
						for( int j = 0; j < results[i] -> numOfWords; j++ )
						{
							encodeString( (results[i] -> words)[j], message + startPointer );
							message[startPointer + (results[i] -> words)[j].length()] = ' ';
							startPointer += (results[i] -> words)[j].length() + 1;
						}
						oss << results[i] -> ngramCount;
						encodeString( oss.str(), message + startPointer );
						message[startPointer + oss.str().length()] = '\n';
						startPointer += oss.str().length() + 1;
						oss.str( "" );
						delete results[i];
					}
	
					break;

				case NO_SMOOTHING_PROB:
					server -> analyzeQueryString( (const char*) stringBuffer, length, qString, language, queryWords );
					it = server -> trees.find( language );
					ts = it -> second;
					probability = ( server -> smoothing ).getProb( qString, queryWords, ts[0] );
					oss << probability;
					messageSize = oss.str().length() + 1;
					message = new byte[messageSize];
					encodeString( oss.str(), message );
					message[messageSize-1] = '\n';

					break;
	
				case ABSOLUTE_DISCOUNTING:
					server -> analyzeQueryString( (const char*) stringBuffer, length, qString, language, queryWords );
					it = server -> trees.find( language );
					ts = it -> second;
					probability = ( server -> smoothing ).getAbsDiscountProb( qString, queryWords, ts[0] );
					oss << probability;
					messageSize = oss.str().length() + 1;
					message = new byte[messageSize];
					encodeString( oss.str(), message );
					message[messageSize-1] = '\n';
	
					break;

				case KN_DISCOUNTING:
					server -> analyzeQueryString( (const char*) stringBuffer, length, qString, language, queryWords );
					it = server -> trees.find( language );
					ts = it -> second;
					probability = ( server -> smoothing ).getKNDiscountProb( qString, queryWords, ts[0], ts[1] );
					oss << probability;
					messageSize = oss.str().length() + 1;
					message = new byte[messageSize];
					encodeString( oss.str(), message );
					message[messageSize-1] = '\n';

					break;

				case KN_MC_DISCOUNTING:
					server -> analyzeQueryString( (const char*) stringBuffer, length, qString, language, queryWords );
					it = server -> trees.find( language );
					ts = it -> second;
					probability = ( server -> smoothing ).getMissingCountsProb( qString, queryWords, ts[0], ts[1] );
					oss << probability;
					messageSize = oss.str().length() + 1;
					message = new byte[messageSize];
					encodeString( oss.str(), message );
					message[messageSize-1] = '\n';

					break;

				case DIRICHLET_DISCOUNTING:
					server -> analyzeQueryString( (const char*) stringBuffer, length, qString, language, queryWords );
					it = server -> trees.find( language );
					ts = it -> second;
					probability = ( server -> smoothing ).getDirichletProb( qString, queryWords, ts[0] );
					oss << probability;
					messageSize = oss.str().length() + 1;
					message = new byte[messageSize];
					encodeString( oss.str(), message );
					message[messageSize-1] = '\n';

					break;

				case DKN_DISCOUNTING:
					server -> analyzeQueryString( (const char*) stringBuffer, length, qString, language, queryWords );
					it = server -> trees.find( language );
					ts = it -> second;
					probability = ( server -> smoothing ).getDKNProb( qString, queryWords, ts[0], ts[1] );
					oss << probability;
					messageSize = oss.str().length() + 1;
					message = new byte[messageSize];
					encodeString( oss.str(), message );
					message[messageSize-1] = '\n';

					break;
			}
		}
		catch( const char* cp )
		{
			perror( cp );

			string error( cp );
			if( error.compare( "recv" ) == 0 )
				break;

			if( message != NULL )
				delete[] message;
			messageSize = 2;
			message = new byte[messageSize];
			message[0] = '0';
			message[1] = '\n';
		}
		pthread_mutex_unlock( &mutex );

		server -> send( message, messageSize, socket );

		if( message != NULL )
			delete[] message;
	}

	pthread_mutex_lock( &mutex );
	server -> closeConnection( socket );
	pthread_mutex_unlock( &mutex );

	pthread_exit( NULL );
}

void readDirectoryStructure( const string& fileName, string& recordDirs, string& indexDirs )
{
	ifstream fo( fileName.c_str() );
	int numOfDirs;
	fo >> numOfDirs;
	for( int i = 0; i < numOfDirs; i++ )
	{
		string dir;
		fo >> dir;
	}
	fo.close();
}

/*****************************************************
**************** MAIN FUNCTION ***********************
******************************************************/
int main( int argc, char* argv[] )
{
	const int port = atoi( argv[1] );
	const int maxConnections = atoi( argv[2] );
	const int backLog = atoi( argv[3] );

	const int numOfLanguages = atoi( argv[4] );

	map<string, vector<MainTree*> > trees;
	int offset = 5;
	try
	{
		cout << "Initializing Server..." << endl;

		for( int i = 0; i < numOfLanguages; i++ )
		{
			const string language( argv[offset] );
			const int minNgram = atoi( argv[offset+1] );
			const int maxNgram = atoi( argv[offset+2] );

			const string originalNgramDir( argv[offset+3] );
			const string indexDir( argv[offset+4] );
			const int isWildcard = atoi( argv[offset+5] );

			offset += 6;
			string resortDir;
			if( isWildcard )
			{
				resortDir = string( argv[offset] );
				offset++;
			}

			const string unigramFile = originalNgramDir + "1gms/vocab";

			cout << '\t' << "Reading keys..." << endl;
			const KeyMap* keyMap = new KeyMap( unigramFile );
			cout << '\t' << "Reading keys...done!" << endl << endl;

			for( int j = 1; j <= maxNgram; j++ )
			{
				const string mainIndexFileName = "roots.index";

				ostringstream oss;
				oss << j;

				const string indDir = indexDir + "Index" + oss.str() + "/"; 
				if( j == 1 )
				{
					trees[language].push_back( new MainTree( minNgram, maxNgram, originalNgramDir, indDir, mainIndexFileName, keyMap ) );
				}
				else
				{
					if( isWildcard )
					{
						const string resortPosDir = resortDir + "ResortPosition" + oss.str() + "/";
						trees[language].push_back( new MainTree( minNgram, maxNgram, resortPosDir, indDir, mainIndexFileName, keyMap ) );
					}
				}
			}
		}

		server = new Server( trees, port, maxConnections, backLog );
		cout << "Initializing Server...Done!" << endl;

		cout << "Server is up and running!" << endl << endl;
		//server -> runInteractive();
		server -> run();
		cout << endl << "Server is going down!" << endl;
	}
	catch( const char* cp )
	{
		cerr << cp << endl;
	}

	return 0;
}
